/*  shuffle - shuffle and print arguments
 *  -------------------------------------
 *  Copyright 2004, Michael J. Pomraning
 *  <mjp\x40pilcrow\x2Emadison\x2Ewi\x2Eus>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: shuffle.c,v 1.1.1.1 2005/01/13 07:19:41 mjp Exp $
 */

#include "config.h"
#define _GNU_SOURCE
#include <stdlib.h>   /* exit, rand */
#include <unistd.h>   /* getpid */
#include <limits.h>   /* INT_MAX */
#include <stdio.h>    /* printf */
#if HAVE_STDBOOL_H
# include <stdbool.h>  /* truth */
#else
# if ! HAVE__BOOL
#  ifdef __cplusplus
typedef bool _Bool;
#  else
typedef unsigned char _Bool;
#  endif
# endif
# define bool _Bool
# define false 0
# define true 1
# define __bool_true_false_are_defined 1
#endif /* HAVE_STDBOOL_H */
#include <errno.h>    /* errno, E* */
#include <time.h>     /* time */
#include <getopt.h>   /* getopt_long */

#define PROG_VERSION    "shuffle 0.1"
static char *progname = NULL;

static bool
str2int(const char *str, int *out) {
  char *endptr = NULL;
  long l;

  if (!str || str[0] == '\0' || !out)
    goto einval;

  errno = 0;
  l = strtol(str, &endptr, 10);
  if (*endptr != '\0') goto einval;
  if ((errno == ERANGE && (l == LONG_MAX || l == LONG_MIN))
      ||
      (l > INT_MAX || l < INT_MIN)) {
    errno = ERANGE; /* may be redundant */
    return false;
  }

  *out = (int) l;

  return true;

einval:
  errno = EINVAL;
  return false;
}

static void
errout(const char *msg) {
  fprintf(stderr, "%s: %s\n", progname, msg);
  exit(EXIT_FAILURE);
}

static void
usage(void) {
  printf(
"Usage: %s [OPTIONS] [--] [ARG]...\n"
"       %s OPTION\n"
"Shuffle and print the ARG(s) to standard output.\n"
"\n"
"  -c, --count=NUM       print only the first NUM ARG(s) after shuffling.\n"
"                        If 0, print all shuffled ARG(s).  Default: 0.\n"
"  -s, --separator=STR   use STR as the output field separator, instead\n"
"                        of a newline.\n"
"      --help            display this help and exit.\n"
"      --version         display version information and exit.\n"
"\n",
    progname, progname);
exit(EXIT_SUCCESS);
}

static void
print_array(char * const a[], const char *sep, size_t sz) {

  if (a == NULL) goto error;

  if (*a == NULL || sz == 0) return; /* Nothing to do */

  while (sz--) {
    if (!*a) goto error;
    fputs(*a++, stdout);
    if (sz > 0) fputs(sep, stdout);
  }
  fputc('\n', stdout);

  return;

error:
  errout("internal error - bad array or array size");
}

static unsigned int
rand_uint_inclusive(unsigned int lo, unsigned int hi) {
  unsigned int delta = hi - lo;

  return
    lo + (unsigned int)((float)(delta+1) * (float)rand()/(RAND_MAX + 1.0));
}

static bool
fy_shuffle(void **array, size_t sz) {

  if (array == NULL) {
    errno = EINVAL;
    return false;
  }

  if (sz < 2) return true;

  if (sz > (RAND_MAX/2)) {
    errno = E2BIG;
    return false;
  }

  while (--sz) {
    void *tmp;
    size_t j;

    j = rand_uint_inclusive(0, sz);
    tmp = array[j];
    array[j] = array[sz];
    array[sz] = tmp;
  }

  return true;
}

int main(int argc, char * const argv[]) {
  int c;
  int count = 0;            /* -c|--count      max no. of args to print */
  char *ofs = "\n";         /* -s|--separator  output field separator   */
  extern char *optarg;
  extern int optind, opterr;
  static struct option long_opts[] = {
    {"help",      no_argument,       0, 'h'},
    {"version",   no_argument,       0, 'v'},
    {"count",     required_argument, 0, 'c'},
    {"separator", required_argument, 0, 's'},
    {0, 0, 0, 0},
  };

  progname = argv[0];

  opterr = 0;
  while ((c = getopt_long(argc, argv, "+c:s:", long_opts, NULL)) != -1) {
    static bool all_done = false;

    switch (c) {
    case 'h':
      usage();
      break; /* Not reached */
    case 'v':
      printf(PROG_VERSION "\n");
      exit(EXIT_SUCCESS);
      break; /* Not reached */
    case 'c':
      if (! str2int(optarg, &count) || count < 0)
        errout("invalid count of items to return");
      break;
    case 's':
      ofs = optarg;
      break;
    default:
      /* Unrecognized option.  Backup and treat as first non-option */
      all_done = true;
      --optind;
      break;
    }
    if (all_done) break;
  }

  argv += optind; argc -= optind;
  if (count <= 0 || count > argc) count = argc; /* How many to print? */

  srand(time(0) + getpid());

  if (! fy_shuffle((void **)argv, argc))
    errout("error: unable to shuffle args");

  print_array(argv, ofs, (size_t) count);

  exit(EXIT_SUCCESS);
}
